/* Copyright (C) 2008, 2009 Stefan Frenz
 * 
 * This file is part of SJC, the Small Java Compiler written by Stefan Frenz.
 * 
 * SJC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SJC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SJC. If not, see <http://www.gnu.org/licenses/>.
 */

package backend.srb;

import compbase.Ops;
import compbase.StdTypes;

import osio.TextPrinter;
import backend.Instruction;

/**
 * SRBI_Single: single SRB instruction 
 * 
 * @author S. Frenz
 * @version 090215 added output of native info, adopted changed SRBTarget
 *  version 090207 added copyright notice
 *  version 090118 added enumerateInstructions and print
 *  version 081226 initial version
 */

public class SRBI_Single extends SRBInstruction {
  public Instruction insDestInTarget;
  
  public SRBI_Single(SRB owner, int maxInstrSize) {
    super(owner, maxInstrSize);
  }
  
  public void resetMark() {
    isDest=false;
  }
  
  public boolean isMarked() {
    return isDest;
  }
  
  public void generateJumpDestsForTarget() {
    SRBI_Single dest;
    if ((dest=(SRBI_Single)jDest)!=null && dest.insDestInTarget==null)
      dest.insDestInTarget=owner.target.getUnlinkedInstruction();
  }
  
  public void convertToNative(int phase) {
    Instruction tInstr;
    int i;
    SRBRegister regObj0=null, regObj1=null, regObj2=null;
    
    SRBTarget target=owner.target;
    if (reg0!=0) regObj0=outer.getReg(reg0);
    if (reg1!=0) regObj1=outer.getReg(reg1);
    if (reg2!=0) regObj2=outer.getReg(reg2);
    if (owner.err) return;
    switch (type) {
      case SRB.I_JUMP:
        if ((tInstr=((SRBI_Single)jDest).insDestInTarget)==null) {
          owner.ctx.out.println("jump with jumpDestInTarget==null");
          owner.err=true;
        }
        else target.handleJump(tInstr, iPar1, phase);
        break;
      case SRB.I_MAGC:
        if (phase==SRB.P_GENERATECODE) {
          tInstr=target.getUnlinkedInstruction();
          for (i=0; i<size; i++) tInstr.code[i]=code[i];
          target.appendInstruction(tInstr);
        }
        break;
      case SRB.I_NONE:
        if (phase==SRB.P_GENERATECODE) {
          if (insDestInTarget!=null) target.appendInstruction(insDestInTarget);
        }
        break;
      case SRB.I_MTHDFRAME:
        target.handleMthdFrame(iPar1, refMthd, phase);
        break;
      case SRB.I_FLOWHINT:
        target.handleFlowHint(iPar1, iPar2, phase);
        break;
      case SRB.I_LOADCONSTVAL:
        target.handleLoadConstVal(regObj0, iPar2, lPar, iPar1, phase);
        break;
      case SRB.I_LOADVAR:
        target.handleLoadVar(iPar3==1, regObj0, regObj1, iPar2, iPar1, phase);
        break;
      case SRB.I_DUPCONV:
        target.handleDupConv(regObj0, regObj1, iPar1, iPar2, iPar3==1, phase);
        break;
      case SRB.I_SINGLESTACKOP:
        target.handleSingleStackOp(iPar2==1, regObj0, iPar3, lPar, iPar1, phase);
        break;
      case SRB.I_ASSIGN:
        target.handleAssign(regObj0, regObj1, iPar1, phase);
        break;
      case SRB.I_REGOP:
        target.handleRegOp(regObj0, regObj1, regObj2, iPar2, iPar1, phase);
        break;
      case SRB.I_MEMOP:
        target.handleMemOp(regObj0, iPar2, iPar1, phase);
        break;
      case SRB.I_CALL:
        target.handleCall(regObj0, iPar1, iPar2, refMthd, iPar3==1, phase);
        break;
      case SRB.I_COMPARE:
        target.handleCompare(regObj0, regObj1, iPar2, iPar3, lPar, iPar1, phase);
        break;
      case SRB.I_SPECIALMEMIO:
        target.handleSpecialMemIO(iPar2==1, regObj0, regObj1, iPar1, iPar3, phase);
        break;
      case SRB.I_DEREFADDR:
        target.handleDerefAddr(regObj0, regObj1, iPar1, iPar2, phase);
        break;
      case SRB.I_CHECKBOUNDS:
        target.handleCheckBounds(regObj0, regObj1, iPar1, jDest, phase);
        break;
      case SRB.I_NATIVEBOUNDEXC:
        target.handleNativeBoundException(0);
        break;
      case SRB.I_THROWFRAME:
        target.handleThrowFrame(iPar1, regObj0, jDest, iPar2, phase);
        break;
      case SRB.I_INLINEVAROFFSET:
        target.handleInlineVarOffset(iPar1, regObj0, iPar2, iPar3, phase);
        break;
    }
  }
  
  public void cleanUp() {
    insDestInTarget=null;
    next=owner.freeInstruction;
    owner.freeInstruction=this;
  }
  
  public int enumerateInstructions(int startWith) {
    return (instrNr=startWith)+1;
  }
  
  public int print(TextPrinter v) {
    int cnt;
    
    if (type==SRB.I_NONE) return 0;
    v.print(instrNr); v.print(": ");
    switch (type) {
      case SRB.I_MAGC:
        v.print("MAGIC");
        for (cnt=0; cnt<size; cnt++) {
          v.print(" 0x");
          v.printHexFix((int)code[cnt], 2);
        }
        v.println();
        return cnt;
      case SRB.I_JUMP:
        v.print("jump ");
        switch (iPar1) { //contains condition
          case 0:        v.print("all"); break; //jump allways
          case Ops.C_LW: v.print("lw"); break; //jump if less
          case Ops.C_LE: v.print("le"); break; //jump if less or equal
          case Ops.C_EQ: v.print("eq"); break; //jump if equal
          case Ops.C_GE: v.print("ge"); break; //jump if greater or equal (==not less)
          case Ops.C_GT: v.print("gt"); break; //jump if greater (==not less or equal)
          case Ops.C_NE: v.print("ne"); break; //jump if not equal
          case Ops.C_BO: v.print("bo"); break; //jump if carry (==below ==not above or equal)
          default: v.print("inv");
        }
        v.print(",i");
        v.print(jDest.instrNr);
        break;
      case SRB.I_MTHDFRAME:
        v.print("mthdframe "); v.print(iPar1); v.print(',');
        if (refMthd!=null) { refMthd.owner.printNameWithOuter(v); v.print('.'); refMthd.printNamePar(v); }
        else v.print("null");
        break;
      case SRB.I_FLOWHINT:
        v.print("flowhint "); v.print(iPar1); v.print(','); v.print(iPar2);
        break;
      case SRB.I_LOADCONSTVAL:
        v.print("loadConstVal r"); v.print(reg0); v.print(','); v.print(iPar2);
        v.print(','); v.print(lPar); v.print(','); StdTypes.printStdType(iPar1, v);
        break;
      case SRB.I_LOADVAR:
        v.print("loadVar "); v.print(iPar3); v.print(",r"); v.print(reg0);
        v.print(",r"); v.print(reg1); v.print(','); v.print(iPar2); v.print(','); StdTypes.printStdType(iPar1, v);
        break;
      case SRB.I_DUPCONV:
        v.print("dupConv r"); v.print(reg0); v.print(",r"); v.print(reg1);
        v.print(','); StdTypes.printStdType(iPar1, v); v.print(','); StdTypes.printStdType(iPar1, v);
        v.print(','); v.print(iPar3);
        break;
      case SRB.I_SINGLESTACKOP:
        v.print("singleStackOp "); v.print(iPar1); v.print(",r"); v.print(reg0);
        v.print(','); v.print(iPar3); v.print(','); v.print(lPar); v.print(','); StdTypes.printStdType(iPar1, v);
        break;
      case SRB.I_ASSIGN:
        v.print("assign r"); v.print(reg0); v.print(",r"); v.print(reg1); v.print(','); StdTypes.printStdType(iPar1, v);
        break;
      case SRB.I_REGOP:
        v.print("regOp r"); v.print(reg0); v.print(",r"); v.print(reg1);
        v.print(',');  v.print(reg2); v.print(','); StdTypes.printStdType(iPar1, v);
        break;
      case SRB.I_MEMOP:
        v.print("memOp r"); v.print(reg0); v.print(','); v.print(iPar2); v.print(','); StdTypes.printStdType(iPar1, v);
        break;
      case SRB.I_CALL:
        v.print("call r"); v.print(reg0); v.print(','); v.print(iPar1); v.print(','); v.print(iPar2);
        v.print(','); refMthd.owner.printNameWithOuter(v); v.print('.'); refMthd.printNamePar(v);
        v.print(','); v.print(iPar3);
        break;
      case SRB.I_COMPARE:
        v.print("compare r"); v.print(reg0); v.print(",r"); v.print(reg1);
        v.print(",r");  v.print(reg2); v.print(','); v.print(iPar2); v.print(','); v.print(iPar3);
        v.print(','); v.print(lPar); v.print(','); StdTypes.printStdType(iPar1, v);
        break;
      case SRB.I_SPECIALMEMIO:
        v.print("specialMemIO "); v.print(iPar2); v.print(",r"); v.print(reg0);
        v.print(",r"); v.print(reg1); v.print(',');  StdTypes.printStdType(iPar1, v); v.print(','); v.print(iPar3);
        break;
      case SRB.I_DEREFADDR:
        v.print("derefAddr r"); v.print(reg0); v.print(",r"); v.print(reg1);
        v.print(',');  v.print(iPar1); v.print(','); v.print(iPar2);
        break;
      case SRB.I_CHECKBOUNDS:
        v.print("checkBounds r"); v.print(reg0); v.print(",r"); v.print(reg1);
        v.print(',');  v.print(iPar1); v.print(",i"); v.print(jDest.instrNr);
        break;
      case SRB.I_NATIVEBOUNDEXC:
        v.print("nativeBoundExc");
        break;
      case SRB.I_THROWFRAME:
        v.print("throwFrame "); v.print(iPar1); v.print(",r"); v.print(reg0);
        v.print(",i"); v.print(jDest.instrNr); v.print(','); v.print(iPar2);
        break;
      case SRB.I_INLINEVAROFFSET:
        v.print("inlineVarOffset "); v.print(iPar1); v.print(",r"); v.print(reg0);
        v.print(','); v.print(iPar2); v.print(','); v.print(iPar3);
        break;
    }
    if (reg0!=0) printNativeReg(reg0, v);
    if (reg1!=0) printNativeReg(reg1, v);
    if (reg2!=0) printNativeReg(reg2, v);
    v.println();
    return 1;
  }
  
  protected void printNativeReg(int srbRegNr, TextPrinter v) {
    SRBRegister reg;
    v.print(" {r");
    v.print(srbRegNr);
    v.print(':');
    if ((reg=outer.getReg(srbRegNr))==null) v.print("!null!");
    else owner.target.printNativeInfo(reg.nativeInfo, v);
    v.print('}');
  }
}
