/* Copyright (C) 2005, 2006, 2007, 2008, 2009 Stefan Frenz
 * 
 * This file is part of SJC, the Small Java Compiler written by Stefan Frenz.
 * 
 * SJC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SJC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SJC. If not, see <http://www.gnu.org/licenses/>.
 */

package compbase;

import backend.Architecture;

/**
 * Vrbl: platform and language independent declaration of a variable 
 * 
 * @author S. Frenz
 * @version 090507 added checks for struct variables minSize to support correct offset calculation of struct-arrays
 *  version 090209 fixed usage of genStoreVarConstVal if type is T_LONG or T_DBL
 *  version 090208 using genStoreVarConstVal instead of genClearMem
 *  version 090207 added copyright notice
 *  version 080802 changed register allocation strategy in genInitCode to reduce register copy after method calls
 *  version 080614 removed copy
 *  version 070917 added optimization to skip zero-initialization
 *  version 070913 added support for genInitCode with init==null
 *  version 070909 optimized signature of Expr.resolve
 *  version 070903 removed checkNameAndType
 *  version 070731 adopted change of renaming id to name
 *  version 070727 added symHint
 *  version 070628 adopted changes for "Expr extends TypeRef"
 *  version 070526 added flag for length of inline-arrays
 *  version 070331 added initialization of movable indirect scalars, fixed location assignment
 *  version 070127 added generic genInitCode
 *  version 070114 added flags to optimize visibility level
 *  version 070111 adapted change in printPos and compErr
 *  version 070106 removed access to inlinedArraySize
 *  version 061203 optimized calls to printPos and compErr
 *  version 061101 added support for structs
 *  version 060607 initial version
 */

public class Vrbl extends AccVar {
  public final static int M_PUB    = Modifier.M_PUB;
  public final static int M_PROT   = Modifier.M_PROT;
  public final static int M_PRIV   = Modifier.M_PRIV;
	public final static int M_FIN    = Modifier.M_FIN;
	public final static int M_STAT   = Modifier.M_STAT;
  public final static int M_STRUCT = Modifier.M_STRUCT; //variable belongs to struct
  public final static int M_ARRLEN = Modifier.M_ARRLEN;
  public final static int MA_ACCSSD = Modifier.MA_ACCSSD;
  public final static int MA_PUB    = Modifier.MA_PUB;
  public final static int MA_PROT   = Modifier.MA_PROT;
	
  public final static String NOREGFREE = "registers not free in Vrbl.genInitCode";
  
	//required fields for resolving
	public Unit owner;
	public int modifier;
	public Vrbl nextVrbl;
  public Expr init;
  public int symHint; //required for symbolinformer
	
	public Vrbl(String ii, int im, int fid, int il, int ic) {
		super(fid, il, ic);
		name=ii;
		modifier=im;
	}
	
	public boolean resolveConstant(Unit owner, Context ctx) {
	  if (location==L_CONST) return true;
	  if (location!=L_CONSTDC) {
      printPos(ctx, "cyclic initialization for variable ");
      ctx.out.print(name);
      return false; //cyclic declaration
    }
	  location=L_CONSTTR;
	  if (init!=null) {
	    if (!init.resolve(owner, owner.initStat, ctx)) {
        ctx.out.print(" in constant initialization for variable ");
        ctx.out.print(name);
	      return false;
      }
      if (init.calcConstantType(ctx)==0 && !init.isCompInitConstObject(ctx)) {
	      init.printPos(ctx, "not constant initialization for variable ");
        ctx.out.print(name);
	      return false;
	    }
      if (init.compareType(type, false, ctx)!=TypeRef.C_EQ) {
        init.printPos(ctx, "constant initialization needs to have exactly the same type of variable ");
        ctx.out.print(name);
        return false;
      }
	  }
	  location=L_CONST;
	  return true;
	}
	
  public boolean enterSize(int loc, Context ctx) {
    //at first, do default things
    if (!super.enterSize(loc, ctx)) return false;
    //do special checks for variables inside a struct
    if ((modifier&M_STRUCT)!=0) {
      if (type.arrDim>0) {
        if (type.arrDim>1) {
          printPos(ctx, "struct-variables must not have array dimension >1");
          return false;
        }
        if (type.baseType!=TypeRef.T_QID) minSize=TypeRef.getMinSize(type.baseType);
      }
      if ((modifier&(M_STAT|M_FIN))!=0) {
        printPos(ctx, "struct-variables can not be \"final\" or \"static\"");
        return false;
      }
      if (type.baseType==TypeRef.T_QID) {
        if (!type.qid.unitDest.resolveInterface(ctx)) return false; //parent must be ok for inlining
        if ((type.qid.unitDest.modifier&Unit.M_STRUCT)==0) {
          printPos(ctx, "struct-variable must not reference non-struct-objects");
          return false;
        }
        minSize=type.qid.unitDest.instScalarTableSize; //enter size of one element if array
      }
      //check if location is to be set
      if (loc==L_UNIT) location=L_STRUCT;
    }
    //otherwise check if variable is declared in a unit and location has to be set
    else if (loc==L_UNIT) {
      if ((modifier&M_FIN)!=0) location=L_CONSTDC;
      else if (minSize<0) {
        if ((modifier&M_STAT)!=0) location=L_CLSSREL;
        else location=L_INSTREL;
      }
      else {
        if ((modifier&M_STAT)!=0) location=L_CLSSSCL;
        else if (ctx.indirScalars) location=L_INSTIDS;
        else location=L_INSTSCL;
      }
    }
    //everything ok
    return true;
  }
  
  public void genInitCode(boolean forceInit, Context ctx) {
    int addr, restore, regType, via, pos;
    Expr toInit;
    
    if ((toInit=init)!=null) { //check if initialization results in zero result
      switch (init.baseType) {
        case StdTypes.T_NULL: case StdTypes.T_NNPT: case StdTypes.T_NDPT: toInit=null; break;
      }
      switch (init.calcConstantType(ctx)) {
        case StdTypes.T_INT: if (init.getConstIntValue(ctx)==0) toInit=null; break;
        case StdTypes.T_LONG: if (init.getConstLongValue(ctx)==0l) toInit=null; break;
      }
    }
    if (toInit==null && !forceInit) return; //no initialization needed for zero values
    regType=type.getRegType(ctx);
    switch (location) {
      case L_INSTREL:
        via=Architecture.M_INST;
        pos=-ctx.arch.relocBytes*varPos;
        break;
      case L_INSTSCL:
        via=Architecture.M_INST;
        pos=varPos;
        break;
      case L_INSTIDS: //very special case: scalar variable addressed indirectly
        restore=ctx.arch.prepareFreeReg(0, 0, 0, StdTypes.T_PTR);
        addr=ctx.arch.allocReg();
        ctx.arch.genLoadVarVal(addr, 0, Architecture.M_INST, ctx.indirScalarAddrOff, StdTypes.T_PTR);
        doGenInit(addr, Architecture.M_RESU, varPos, regType, toInit, ctx);
        ctx.arch.deallocRestoreReg(addr, 0, restore);
        return;
      case L_LOCAL:
        via=Architecture.M_STCK;
        pos=varPos;
        break;
      default:
        compErr(ctx, "can not initialize variable beside instance or local");
        return;
    }
    doGenInit(0, via, pos, regType, toInit, ctx);
  }
  
  private void doGenInit(int resReg, int via, int pos, int regType, Expr toInit, Context ctx) {
    int value;
    boolean asDouble=false;
    
    if (toInit!=null) switch (init.calcConstantType(ctx)) { //value-init
      case StdTypes.T_INT: //constant 8/16/32 bit value
        ctx.arch.genStoreVarConstVal(resReg, init.getConstIntValue(ctx), via, pos, regType);
        break;
      case StdTypes.T_LONG: //constant 64 bit value
        ctx.arch.genStoreVarConstDoubleOrLongVal(resReg, init.getConstLongValue(ctx), via, pos, regType==StdTypes.T_DBL);
        break;
      default: //not a constant value
        if (ctx.arch.prepareFreeReg(0, 0, 0, regType)!=0) {
          compErr(ctx, NOREGFREE);
          return;
        }
        value=ctx.arch.allocReg(); //address of variable does not change, value is generatable already
        init.genOutputVal(value, ctx);
        ctx.arch.genStoreVarVal(resReg, value, via, pos, regType);
        ctx.arch.deallocRestoreReg(value, 0, 0);
    }
    else { //clear memory
      switch (regType) {
        case StdTypes.T_DBL:
          asDouble=true; //no break
        case StdTypes.T_LONG:
          ctx.arch.genStoreVarConstDoubleOrLongVal(resReg, 0l, via, pos, asDouble);
          break;
        default:
          ctx.arch.genStoreVarConstVal(resReg, 0, via, pos, regType);
      }
    }
  }
}
