/* Copyright (C) 2005, 2006, 2007, 2008, 2009 Stefan Frenz
 * 
 * This file is part of SJC, the Small Java Compiler written by Stefan Frenz.
 * 
 * SJC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SJC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SJC. If not, see <http://www.gnu.org/licenses/>.
 */

package frontend.sjava;

import compbase.*;

/**
 * ExVar: access to a variable
 * 
 * @author S. Frenz
 * @version 090508 fixed unit search order 1) imported units, 2) local package, 3) imported packages
 *  version 090221 changed all access level checks to centralized version, fixed outerLevel calculation
 *  version 090218 made use of centralized access level check
 *  version 090207 added copyright notice
 *  version 090206 moved isAddrOnStack to ExAccVrbl
 *  version 080911 fixed access to other inner class in outer class through name of class
 *  version 080813 fixed access to not resolved variables in constant resolving
 *  version 080616 added check for invalid constant objects
 *  version 080614 adopted merging of T_CLSS/T_INTF in StdTypes, adopted changed Unit.searchVariable
 *  version 080610 removed conversion to JMthd as it is no longer required
 *  version 080523 fixed error message for not possible access of constant 
 *  version 080306 fixed too restrictive package-check of bugfix from 080305
 *  version 080305 fixed invalid package-match if no package result possible
 *  version 080220 varResolve limits search if resolving deref
 *  version 080105 adopted changed signature of Unit.isParent
 *  version 071227 added support for inner units
 *  version 070913 adopted change in Mthd
 *  version 070909 optimized signature of Expr.resolve, adopted changed ExAccVrbl
 *  version 070813 fixed unit context during resolving of static variable in dynaMem mode
 *  version 070809 fixed owner during resolving of foreign variable
 *  version 070731 adopted change of QualID[] to QualIDList
 *  version 070727 replaced exSubResolve by resolve as there is nothing done anymore
 *  version 070628 adopted changes for "Expr extends TypeRef"
 *  version 070114 reduced access level where possible
 *  version 070111 adapted change in printPos and compErr
 *  version 070106 moved array stuff to ExDeArray
 *  version 070104 fixed addrIsOnStack
 *  version 061203 optimized calls to printPos and compErr
 *  version 061202 optimized static modes
 *  version 061129 static TypeRef object moved dynamically to Context
 *  version 061127 added support for embedded mode
 *  version 061111 added checks for invalid struct-array-derefs
 *  version 061107 added support for indexed struct-variables
 *  version 061105 added support for variables inside a STRUCT
 *  version 061030 changed detection of indirectCall
 *  version 061027 optimized check for special names
 *  version 060804 added ensureClassContext
 *  version 060803 added check for ownership of static variables
 *  version 060714 changed bound/null check to support runtime handler
 *  version 060705 changed deref-execution order (now Java-standard)
 *  version 060628 added support for static compilation
 *  version 060620 bugfix long-int
 *  version 060619 bugfix for register-allocation in genCondJmp
 *  version 060607 initial version
 */

public class ExVar extends ExAccVrbl {
	protected PureID id;
  
	protected ExVar(PureID ip, int fid, int il, int ic) {
		super(fid, il, ic);
		id=ip;
	}
	
	public boolean resolve(Unit unitContext, Mthd mthdContext, Context ctx) {
    if (id.name.equals("this")) return varResolveThis(unitContext, mthdContext, ctx);
	  if (!varResolve(null, unitContext, mthdContext, unitContext, mthdContext, false, ctx)) return false;
	  if (dest==null) { //destination is class or package
	    printPos(ctx, "var ");
	    ctx.out.print(id.name);
	    ctx.out.print(" not found");
	    return false;
    }
	  return true;
	}
  
  protected boolean varResolveThis(Unit inUnit, Mthd inMthd, Context ctx) {
    if ((inMthd.modifier&JMthd.M_STAT)!=0) {
      printPos(ctx, "\"this\" can not be used in static methods");
      return false;
    }
    baseType=T_QID;
    qid=inUnit.getTypeTo().qid;
    isThis=true;
    return true;
  }
  
  protected boolean varResolveSuper(Unit inUnit, Mthd inMthd, Context ctx) {
    Unit parent;
    if (inUnit==ctx.langRoot.unitDest) {
      printPos(ctx, "\"super\" is not possible in lang-root");
      return false;
    }
    if (!(inUnit instanceof Clss)) {
      printPos(ctx, "\"super\" is only possible for classes");
      return false;
    }
    if ((inMthd.modifier&Mthd.M_STAT)!=0) {
      printPos(ctx, "\"super\" not allowed in static context");
    }
    parent=((Clss)inUnit).extsID.unitDest;
    baseType=T_QID;
    qid=parent.getTypeTo().qid;
    if (ctx.dynaMem) importedClass=inUnit.getRefUnit(parent, true);
    return true;
  }
  
	//has also to resolve units and packets for ExDeRef
	protected boolean varResolve(Pack inPack, Unit inUnit, Mthd inMthd, Unit unitContext, Mthd mthdContext,
	    boolean asDeRef, Context ctx) {
	  Param vParam=null;
	  Vrbl vLocal=null, dvar;
    QualIDList impt=null;
    
	  //check for this, super, MAGIC and MARKER
	  if (SJava.isSpecialName(id.name)) {
	    printPos(ctx, "keyword used in improper way");
	    return false;
	  }
	  //search name
	  if (inMthd!=null) {
	    vParam=inMthd.param;
	    vLocal=inMthd.vars;
	  }
    //else: already initialized with null
    if (inUnit instanceof JUnit) impt=((JUnit)inUnit).impt;
	  if (asDeRef || (!resolveInVar(vLocal, ctx) && !resolveInParam(vParam, ctx))) {
      if (inUnit!=null && inUnit.innerUnits!=null && resolveInInner(inUnit.innerUnits, ctx)) { //check inner units
        return minimumAccessLevel(id.name, unitContext, qid.unitDest, qid.unitDest.modifier, ctx)!=L_NOTVISIBLE;
      }
      //access not through inner class
      if (inUnit==null || !resolveInUnit(inUnit, ctx)) {
        if (!resolveInPackage(inPack, true) && (asDeRef || (inUnit==null
            || (!resolveInImptList(impt, false) && !resolveInPackage(inUnit.pack.packDest, asDeRef)
                && !resolveInImptList(impt, true) && !resolveInPackage(ctx.root, true)
                && !resolveInPackage(ctx.defUnits, false) && !resolveInPackage(ctx.rte, false))))) {
          if (inUnit!=null && inUnit.outerUnit!=null) { //try to resolve in outer unit
            if (varResolve(inPack, inUnit.outerUnit, inMthd, unitContext, mthdContext, asDeRef, ctx)) {
              //access of variable
              if (!(dest instanceof Vrbl)) {
                compErr(ctx, "var-dest in outer is not of type Vrbl");
                return false;
              }
              dvar=(Vrbl)dest;
              if ((dvar.modifier&Vrbl.M_STAT)==0) { //access via instance
                if ((inUnit.modifier&Unit.M_STAT)!=0) {
                  printPos(ctx, "can not access instance variable of outer unit in static unit");
                  return false;
                }
                outerLevel++; //increase level
                outerAccessStart=inUnit; //last write == outest unit wins
              }
              else { //access via class descriptor
                ensureClassContext=true;
                if (ctx.dynaMem) importedClass=unitContext.getRefUnit(dvar.owner, true);
              }
              return true; //found variable in outer unit
            }
            return false; //error message already printed
          }
          //variable not found
    	    printPos(ctx, "identifier ");
    	    ctx.out.print(id.name);
    	    ctx.out.print(" not found");
    	    return false;
        }
        //check access level of class or interface
        if (baseType!=StdTypes.T_PACK && minimumAccessLevel(id.name, unitContext, 
            qid.unitDest, qid.unitDest.modifier, ctx)==L_NOTVISIBLE) return false;
  	  }
  	  else { //check unit-variable
  	    dvar=(Vrbl)dest;
        if (inMthd==null || dvar==null) {
          compErr(ctx, "inMthd/dvar==null");
          return false;
        }
        //check if owner is resolved already
        if (dvar.location==AccVar.L_NOTRDY && !dvar.owner.resolveInterface(ctx)) return false; //output already done
        //check if constant
        if (dvar.location==AccVar.L_CONSTDC || dvar.location==AccVar.L_CONSTTR
            || dvar.location==AccVar.L_CONST) {
          if (dvar.location==AccVar.L_CONSTDC && !dvar.resolveConstant(dvar.owner, ctx))
            return false; //output already done
          if (dvar.location!=AccVar.L_CONST) {
            printPos(ctx, "unresolvable or cyclic initialization of variable");
            return false;
          }
          if (dvar.init!=null && (constType=dvar.init.calcConstantType(ctx))==0 &&
              !(constObject=dvar.init.isCompInitConstObject(ctx))) {
            printPos(ctx, "unknown type of constant variable");
            return false;
          }
          if (constObject && (dest=dvar=dvar.init.getCompInitConstObjectVrbl(ctx))==null) {
            compErr(ctx, "invalid constant object");
            return false;
          }
        }
        //check static access
    	  if ((dvar.modifier&Vrbl.M_STAT)==0) {
          if ((inMthd.modifier&Mthd.M_STAT)!=0) {
      	    printPos(ctx, "can not access dynamic variable in static context");
      	    return false;
          }
    	  }
        else {
          if ((inUnit!=dvar.owner || (inMthd.modifier&Mthd.M_STAT)==0)
              && (constObject || (dvar.modifier&Vrbl.M_FIN)==0)) {
            ensureClassContext=true;
            qid=dvar.owner.getTypeTo().qid;
            if (ctx.dynaMem) importedClass=unitContext.getRefUnit(dvar.owner, true);
          }
        }
        //check access level
    	  switch (minimumAccessLevel(dvar.name, unitContext, dvar.owner, dvar.modifier, ctx)) {
    	    case L_NOTVISIBLE:
    	      return false;
    	    case L_PROTECTED:
    	      dvar.modifier|=Vrbl.MA_PROT;
    	      break;
    	    case L_PUBLIC:
    	      dvar.modifier|=Vrbl.MA_PUB;
    	      break;
    	  }
        //everything ok
        dvar.modifier|=Vrbl.MA_ACCSSD;
        getTypeOf(dvar.type);
  	  }
	  }
	  //else: local and parameter variable need not to be checked for modifiers
	  //everything OK
	  return true;
	}
	
  public int getConstIntValue(Context ctx) {
    Vrbl dvar;
    
    if (dest instanceof Vrbl) {
      dvar=(Vrbl)dest;
  	  if (dvar.init==null) {
  	    dest.printPos(ctx, "dest.init==null (int)");
        ctx.err=true;
  	    return 0;
  	  }
  	  return dvar.init.getConstIntValue(ctx);
    }
    dest.printPos(ctx, "dest not Vrbl (int)");
    ctx.err=true;
    return 0;
	}
	
	public long getConstLongValue(Context ctx) {
    Vrbl dvar;
    
    if (dest instanceof Vrbl) {
      dvar=(Vrbl)dest;
      if (dvar.init==null) {
        dest.printPos(ctx, "dest.init==null (long)");
        ctx.err=true;
        return 0l;
      }
  	  return dvar.init.getConstLongValue(ctx);
    }
    dest.printPos(ctx, "dest not Vrbl (long)");
    ctx.err=true;
    return 0l;
	}
  
  public Vrbl getCompInitConstObjectVrbl(Context ctx) {
    if (constObject) return (Vrbl)dest;
    return null;
  }
  
	private boolean resolveInVar(Vrbl vars, Context ctx) {
	  while (vars!=null) {
	    if (id.name.equals(vars.name)) {
	      dest=vars;
	      getTypeOf(dest.type); 
	      return true;
	    }
	    vars=vars.nextVrbl;
	  }
	  return false;
	}

	private boolean resolveInParam(Param vars, Context ctx) {
	  while (vars!=null) {
	    if (id.name.equals(vars.name)) {
	      dest=vars;
	      getTypeOf(dest.type); 
	      return true;
	    }
	    vars=vars.nextParam;
	  }
	  return false;
	}
	
  private boolean resolveInInner(Unit unit, Context ctx) {
    while (unit!=null) {
      if (id.name.equals(unit.name)) {
        qid=unit.getTypeTo().qid;
        baseType=StdTypes.T_DESC;
        return true;
      }
      if (resolveInInner(unit.innerUnits, ctx)) return true;
      unit=unit.nextUnit;
    }
    return false;
  }
  
	private boolean resolveInUnit(Unit unit, Context ctx) {
	  if ((dest=unit.searchVariable(id.name, ctx))==null) return false;
    qid=unit.getTypeTo().qid;
	  getTypeOf(dest.type);
	  return true;
	}

	private boolean resolveInPackage(Pack pack, boolean allowPacks) {
	  Pack packDest;
	  Unit unitDest;
	  
	  if (pack==null) return false;
	  if (allowPacks && (packDest=pack.searchSubPackage(id.name))!=null) {
      qid=packDest.getTypeTo().qid;
	    baseType=StdTypes.T_PACK;
	    return true;
	  }
	  if ((unitDest=pack.searchUnit(id.name))!=null) {
	    qid=unitDest.getTypeTo().qid;
	    baseType=StdTypes.T_DESC;
	    return true;
	  }
	  return false;
	}
	
	private boolean resolveInImptList(QualIDList impt, boolean packagesNotUnits) {
	  while (impt!=null) {
	    if (packagesNotUnits) {
	      if (impt.qid.type==QualID.Q_IMPORTPACK) {
	        if (resolveInPackage(impt.qid.packDest, true)) return true;
	      }
	    }
	    else {
	      if  (impt.qid.type==QualID.Q_IMPORTUNIT) {
	        if (impt.qid.unitDest.name.equals(id.name)) {
	          qid=impt.qid;
	          baseType=StdTypes.T_DESC;
	          return true;
	        }
	      }
	    }
      impt=impt.nextQualID;
    }
	  return false;
	}
	
	public void genOutput(Context ctx) { //expression that consists only of this ExVar -> value is not of interest
    compErr(ctx, "ExVar.genOutput called which has no effect");
	}
}
