/* Copyright (C) 2005, 2006, 2007, 2008, 2009 Stefan Frenz
 * 
 * This file is part of SJC, the Small Java Compiler written by Stefan Frenz.
 * 
 * SJC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SJC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SJC. If not, see <http://www.gnu.org/licenses/>.
 */

package frontend.sjava;

import compbase.*;

/**
 * StReturn: return-statement
 * 
 * @author S. Frenz
 * @version 090508 adopted changes in Stmt
 *  version 090507 added flow analysis information
 *  version 090207 added copyright notice
 *  version 080711 fixed insertion of implicit conversion for interface return value
 *  version 070909 optimized signature of Expr.resolve, adopted StBlock extends StBreakable
 *  version 070908 optimized signature of Stmt.resolve
 *  version 070628 adopted changes for "Expr extends TypeRef"
 *  version 070114 reduced access level where possible
 *  version 070111 adapted change in printPos and compErr
 *  version 061229 removed access to firstInstr
 *  version 061203 optimized calls to printPos and compErr
 *  version 061111 minor changes
 *  version 060620 added test for implicit conversion
 *  version 060607 initial version
 */

public class StReturn extends Stmt {
	protected Expr retVal;
	private StBlock mthdBlock;
	
	protected StReturn(int fid, int il, int ic) {
		super(fid, il, ic);
	}
	
	protected int innerResolve(int flowCode, Unit unitContext, Mthd mthdContext, Context ctx) {
	  int res;
	  ExEnc enc; //enclosure for implicit conversion
	  
	  if (retVal!=null) { //the programmer wants us to return something, check type
	    if (mthdContext.isConstructor) {
	      printPos(ctx, "can not return anything in constructor");
	      return FA_ERROR;
	    }
	    if (!retVal.resolve(unitContext, mthdContext, ctx)) return FA_ERROR;
      if (retVal.baseType==StdTypes.T_NULL) {
        res=mthdContext.retType.getRegType(ctx);
        if (res==StdTypes.T_PTR) retVal.baseType=StdTypes.T_NNPT;
        else if (res==StdTypes.T_DPTR) retVal.baseType=StdTypes.T_NDPT;
        else {
          printPos(ctx, "null does not match declared type (");
          mthdContext.retType.printType(ctx.out);
          ctx.out.print(")");
          return FA_ERROR;
        }
      }
      else {
        res=retVal.compareType(mthdContext.retType, true, ctx); //get type-compare
        if (res==TypeRef.C_NP || res==TypeRef.C_TT) {
          printPos(ctx, "type of return-expression (");
          retVal.printType(ctx.out);
          ctx.out.print(") does not match declared type (");
          mthdContext.retType.printType(ctx.out);
          ctx.out.print(")");
          return FA_ERROR;
        }
        if (res==TypeRef.C_OT && mthdContext.retType.isIntfType()) { //insert implicit conversion of object to interface
          enc=new ExEnc(fileID, line, col);
          enc.convertTo=mthdContext.retType;
          enc.ex=retVal;
          if (!enc.implConvResolve(unitContext, ctx)) {
            ctx.out.print(" in inserted implicit conversion");
            return FA_ERROR;
          }
          retVal=enc;
        }
      }
	  }
	  else { //the programmer wants to return nothing, check context
	    if (!mthdContext.isConstructor && mthdContext.retType.baseType!=TypeRef.T_VOID) {
        printPos(ctx, "method needs to return something (");
        mthdContext.retType.printType(ctx.out);
        ctx.out.print(")");
        return FA_ERROR;
	    }
	  }
    if (!(mthdContext instanceof JMthd)) {
      printPos(ctx, "java statement in not-java method is not supported");
      return FA_ERROR;
    }
	  mthdBlock=((JMthd)mthdContext).block;
	  return flowCode|FA_NEXT_IS_UNREACHABLE;
	}
	
	protected void genOutput(Context ctx) {
	  int reg, type;
	  
    if (retVal!=null) {
      type=retVal.getRegType(ctx);
  	  if (ctx.arch.prepareFreeReg(0, 0, 0, type)!=0) { //there must be free registers
  	    compErr(ctx, "no free reg at beginning of return-statement");
  	    return;
  	  }
  	  reg=ctx.arch.allocReg();
  	  retVal.genOutputVal(reg, ctx);
      ctx.arch.genMoveToPrimary(reg, type);
  	  ctx.arch.deallocRestoreReg(reg, 0, 0);
  	}
	  ctx.arch.genJmp(mthdBlock.breakDest);
	}
}
