/* Copyright (C) 2009 Stefan Frenz
 * 
 * This file is part of SJC, the Small Java Compiler written by Stefan Frenz.
 * 
 * SJC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SJC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SJC. If not, see <http://www.gnu.org/licenses/>.
 */

package frontend.sjava;

import backend.Architecture;
import compbase.*;

/**
 * StSync: synchronized-statement
 * 
 * @author S. Frenz
 * @version 090508 adopted changes in Stmt
 *  version 090506 added flow analysis information
 *  version 090218 initial version
 */

public class StSync extends Stmt {
  public final static int SYNC_NORM = 0; //normal sync block
  public final static int SYNC_INST = 1; //sync on instance
  public final static int SYNC_CLSS = 2; //sync on class context
  
  protected Expr syncObj;
  protected StBlock syncBlock;
  private UnitList runtimeClass;
  private int syncType, syncObjOffset;
	
  protected StSync(int is, int fid, int il, int ic) {
    super(fid, il, ic);
    syncType=is;
  }
	
	protected int innerResolve(int flowCode, Unit unitContext, Mthd mthdContext, Context ctx) {
	  //resolve syncBlock
	  if (((flowCode=syncBlock.resolve(flowCode, unitContext, mthdContext, ctx))&FA_ERROR)!=0) return FA_ERROR;
	  //resolve syncObj if explicitly given
    if (syncType==SYNC_NORM) {
      if (!syncObj.resolve(unitContext, mthdContext, ctx)) return FA_ERROR;
  	  if (!syncObj.isObjType()) {
  	    printPos(ctx, "need object to synchronize on");
        return FA_ERROR;
  	  }
  	  if (syncObj.baseType==StdTypes.T_NNPT) {
  	    printPos(ctx, "null is not a valid object to synchronize on");
  	    return FA_ERROR;
  	  }
  	  //allocate space for buffer variable if synchronization is not done on inst/clss
  	  mthdContext.varSize=-(syncObjOffset
  	    =(-mthdContext.varSize-ctx.arch.relocBytes)&~ctx.arch.stackClearBits);
	  }
	  //everything ok
	  if (ctx.dynaMem) runtimeClass=unitContext.getRefUnit(ctx.rteDynamicRuntime, true);
	  return flowCode;
	}
	
	protected void genOutput(Context ctx) {
	  int regSyncObj;
	  //evaluate expression and call runtime for "enter"
    if (syncType==SYNC_NORM) {
      if (ctx.arch.prepareFreeReg(0, 0, 0, StdTypes.T_PTR)!=0) { //should be 0 always
        compErr(ctx, "no free reg at beginning of synchronized");
        return;
      }
      regSyncObj=ctx.arch.allocReg();
      syncObj.genOutputVal(regSyncObj, ctx);
      ctx.arch.genStoreVarVal(0, regSyncObj, Architecture.M_STCK, syncObjOffset, StdTypes.T_PTR);
      genOutputDoSyncCall(regSyncObj, 1, ctx);
  	  ctx.arch.deallocRestoreReg(regSyncObj, 0, 0);
    }
    else genOutputDoSyncCall(ctx.arch.getClssInstReg(syncType==SYNC_INST), 1, ctx);
	  //generate code for synchronized block
	  syncBlock.genOutput(ctx);
    //get evaluated expression and call runtime for "leave"
	  if (syncType==SYNC_NORM) {
      if (ctx.arch.prepareFreeReg(0, 0, 0, StdTypes.T_PTR)!=0) { //should be 0 always
        compErr(ctx, "no free reg at beginning of synchronized");
        return;
      }
      regSyncObj=ctx.arch.allocReg();
      ctx.arch.genLoadVarVal(regSyncObj, 0, Architecture.M_STCK, syncObjOffset, StdTypes.T_PTR);
      genOutputDoSyncCall(regSyncObj, 0, ctx);
      ctx.arch.deallocRestoreReg(regSyncObj, 0, 0);
	  }
	  else genOutputDoSyncCall(ctx.arch.getClssInstReg(syncType==SYNC_INST), 0, ctx);
	}
	
	private void genOutputDoSyncCall(int regSyncObj, int enter, Context ctx) {
    if (ctx.dynaMem) ctx.arch.genSaveUnitContext();
    ctx.arch.genPush(regSyncObj, StdTypes.T_PTR); //push sync obj
    ctx.arch.genPushConstVal(enter, StdTypes.T_BOOL); //push "enter"
    if (ctx.dynaMem) {
      ctx.arch.genLoadUnitContext(0, -ctx.arch.relocBytes*runtimeClass.index, true);
      ctx.arch.genCall(-ctx.arch.relocBytes*ctx.rteDoSyncMd.jumpTableIndex, 0, true, ctx.rteDoSyncMd.parSize);
      ctx.arch.genRestUnitContext();
    }
    else ctx.arch.genCallConst(ctx.rteDoSyncMd, ctx.rteDoSyncMd.parSize);
	}
}
