/* Copyright (C) 2006, 2007, 2008, 2009 Stefan Frenz
 * 
 * This file is part of SJC, the Small Java Compiler written by Stefan Frenz.
 * 
 * SJC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SJC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SJC. If not, see <http://www.gnu.org/licenses/>.
 */

package memory;

import compbase.Context;
import compbase.Unit;

/**
 * BootableImage: container of a to be compiled image
 * 
 * @author S. Frenz
 * @version 090207 added copyright notice
 *  version 080202 added support for noSizeScalars-flag
 *  version 070920 fixed bug in finalizeImage
 *  version 070918 added support for multidimensional arrays
 *  version 070729 added image block size check for indirect scalar allocation
 *  version 070703 replaced statistic method by a more versatile one
 *  version 070628 changed stackClearBits to allocClearBits
 *  version 070615 added method to request the indirect part of the last allocated object
 *  version 070528 added statistic method
 *  version 070526 adopted removal of Context.rteSArray* short-paths
 *  version 070506 added support for runtime environment without objects
 *  version 070504 added support for self-allocation of ctx.rteSClassDesc without backfixing
 *  version 070331 added support for additional indirect scalars in SArray
 *  version 070127 optimized access to err-flag
 *  version 061222 option to have header replaced
 *  version 061220 fixed entering of type.outputAddr
 *  version 061217 made header optional
 *  version 061128 added enterInitObject
 *  version 060902 moved to memory package to support bootstrapped compilation
 *  version 060817 removed abstraction to make compression easier
 *  version 060817 added version history
 */

public class BootableImage extends ImageContainer {
  private boolean memBad, standardHeader;
  private int firstObj, lastObj, lastIndirAddr;
  private Context ctx;
  
  //memory block contains information about the image (fixed by loader, writer, system... all!)
  // off= 0: 32 Bit: ptr: start in memory
  // off= 4: 32 Bit: int: size of memory block
  // off= 8: 32 Bit: ptr: class descriptor address of method to start
  // off=12: 32 Bit: ptr: address of first code byte in method to start
  // off=16: 32 Bit: ptr: address of first object
  // off=20: 32 Bit: ptr: address of RAM-init object if in embedded mode
  // off=24: 32 Bit: int: codeStart
  // off=28: 32 Bit: int: lo to hi: 0xAA relocBytes stackClearBits 0x55 (arch parameters incl. endianess)
  
  public void init(int size, int base, boolean iSH, byte[] headerData, Context ictx) {
    int i;
    
    ctx=ictx;
    memBlock=new byte[size];
    baseAddress=base; //this is the offset to which the image should be loaded to
    firstObj=lastObj=-1;
    if (standardHeader=iSH) {
      memBlockLen=(32+ctx.arch.allocClearBits)&~ctx.arch.allocClearBits; //reserve memory for info above
      ctx.arch.putRef(base, base);
      ctx.arch.putInt(base+28, 0x550000AA|(ctx.arch.allocClearBits<<16)|(ctx.arch.relocBytes<<8));
    }
    else {
      if (headerData!=null) {
        memBlockLen=headerData.length;
        for (i=0; i<memBlockLen; i++) ctx.mem.putByte(base+i, headerData[i]);
      }
      else memBlockLen=0;
    }
  }
  
  public void finalizeImage(int unitAddr, int mthdCodeAddr, int relCodeStart) {
    startUnit=unitAddr;
    startCode=mthdCodeAddr;
    if (standardHeader) {
      ctx.arch.putInt(baseAddress+4, memBlockLen); //enter used size
      ctx.arch.putRef(baseAddress+8, startUnit); //enter and remember startup-unit
      ctx.arch.putRef(baseAddress+12, startCode); //enter and remember address to exec
      ctx.arch.putRef(baseAddress+16, firstObj); //enter address of first object
      ctx.arch.putInt(baseAddress+24, relCodeStart); //enter relative start of code
    }
  }
  
  public byte[] getInfoBlock(String appendSig) {
    byte[] data;
    int i, appLen=0;
    
    if (appendSig!=null) appLen=appendSig.length();
    data=new byte[32+appLen];
    for (i=0; i<32; i++) data[i]=memBlock[i];
    for (i=0; i<appLen; i++) data[i+32]=(byte)appendSig.charAt(i);
    return data;
  }
  
  public int allocate(int scalarSize, int indirScalarSize, int relocEntries, int typeAddr) {
    int relocSize, ret, size;
    
    if (memBad) return -1;
    relocSize=relocEntries*ctx.arch.relocBytes;
    size=(relocSize+scalarSize+ctx.arch.allocClearBits)&~ctx.arch.allocClearBits;
    if (memBlockLen+size>memBlock.length) { //no space left
      memError();
      return -1;
    }
    ret=memBlockLen+relocSize+baseAddress; //calculate address
    if (!streamObjects) {
      if (lastObj!=-1) ctx.arch.putRef(lastObj-2*ctx.arch.relocBytes, ret); //enter in next-pointer of last object
      else firstObj=ret;
      if (!noSizeScalars) {
        ctx.arch.putInt(ret, relocEntries); //remember relocSize on first scalar
        ctx.arch.putInt(ret+4, scalarSize); //remember scalarSize on second scalar
      }
      lastObj=ret; //remember for next allocate
    }
    memBlockLen+=size;
    if (indirScalarSize!=0) {
      indirScalarSize=(indirScalarSize+ctx.arch.allocClearBits)&~ctx.arch.allocClearBits;
      if (memBlockLen+indirScalarSize>memBlock.length) { //no space left
        memError();
        return -1;
      }
      ctx.arch.putInt(ret+ctx.indirScalarSizeOff, indirScalarSize); //remember indirScalarSize
      ctx.arch.putRef(ret+ctx.indirScalarAddrOff, lastIndirAddr=baseAddress+memBlockLen); //remember indirScalarAddr
      memBlockLen+=indirScalarSize;
    }
    //enter type
    switch (typeAddr) {
      case -1: break;
      case 0: ctx.arch.putRef(ret-ctx.arch.relocBytes, ret); break;
      default: ctx.arch.putRef(ret-ctx.arch.relocBytes, typeAddr);
    }
    //statistics
    objectsAllocated++;
    return ret;
  }
  
  private void memError() {
    ctx.out.println("no space left in memory block");
    memBad=ctx.err=true;
  }
  
  public int allocateArray(int entries, int dim, int stdEntrySize, int stdType, int extTypeAddr) {
    int obj=-1;
    Unit array=ctx.rteSArray;
    
    if (dim>1 || extTypeAddr!=0) {
      obj=allocate(
        array.instScalarTableSize, array.instIndirScalarTableSize,
        array.instRelocTableEntries+entries*ctx.arch.relocBytes, array.outputAddr);
    }
    else if (stdType!=0) {
      if (ctx.indirScalars) obj=allocate(
          array.instScalarTableSize, array.instIndirScalarTableSize+entries*stdEntrySize,
          array.instRelocTableEntries, array.outputAddr);
      else obj=allocate(
          array.instScalarTableSize+entries*stdEntrySize, 0,
          array.instRelocTableEntries, array.outputAddr);
      //ctx.arch.putRef(obj-ctx.rteSArrayExt*ctx.arch.relocBytes, 0);
    }
    else return -1; //either stdType or extType has to be non-zero
    if (obj<0) return -1; //allocation failed
    ctx.arch.putInt(obj+ctx.rteSArrayLength, entries);
    ctx.arch.putInt(obj+ctx.rteSArrayDim, dim);
    ctx.arch.putInt(obj+ctx.rteSArrayStd, stdType);
    ctx.arch.putRef(obj-ctx.arch.relocBytes*ctx.rteSArrayExt, extTypeAddr);
    return obj;
  }
  
  public void enterInitObject(int objectAddress) {
    if (standardHeader) ctx.arch.putRef(baseAddress+20, objectAddress);
  }
  
  public int getIndirAddrOfLastObject() {
    return lastIndirAddr;
  }
  
  public int getCurrentAllocAmountHint() {
    return memBlockLen;
  }
}
